import React from 'react';
import {
  View,
  Platform,
  PanResponder,
  StyleSheet,
  I18nManager,
  Animated,
  Easing,
  Image,
  Dimensions,
  StatusBar,
} from 'react-native';
import { CardStack as ReactNavigationCardStack, NavigationActions } from 'react-navigation';
import clamp from 'clamp';
import GridWebView from 'grid-react-native-webview';
import clientJsProvider from 'grid-clientjs-provider';
import { get } from 'lodash';
import onNavigationStateChange from './on-navigation-state-change';
import Header from './header';
import WebViewActions from '../routers/webview-actions';

import ReactNavigationHeader from 'react-navigation/src/views/Header';

const styles = StyleSheet.create({
  container: {
    flex: 1,
    // Header is physically rendered after scenes so that Header won't be
    // covered by the shadows of the scenes.
    // That said, we'd have use `flexDirection: 'column-reverse'` to move
    // Header above the scenes.
    flexDirection: 'column-reverse',
  },
  scenes: {
    flex: 1,
  },
});

// The max duration of the card animation in milliseconds after released gesture.
// The actual duration should be always less then that because the rest distance
// is always less then the full distance of the layout.
const ANIMATION_DURATION = 250;

// The gesture distance threshold to trigger the back behavior. For instance,
// `1 / 3` means that moving greater than 1 / 3 of the width of the screen will
// trigger a back action
const POSITION_THRESHOLD = 1 / 3;

// The threshold (in pixels) to start the gesture action.
const RESPOND_THRESHOLD = 12;

// The distance of touch start from the edge of the screen where the gesture will be recognized
const GESTURE_RESPONSE_DISTANCE_HORIZONTAL = 35;
const GESTURE_RESPONSE_DISTANCE_VERTICAL = 135;

class CardStack extends ReactNavigationCardStack {
  fakeHeaderHeight = new Animated.Value(0);
  pageLoading = true;

  componentDidMount() {
    // NOTE: This will overwrite any default componentDidMount set by ReactNavigationCardStack
    this.props.navigation.dispatch(WebViewActions.ref({
      ref: this.webView,
      key: this.props.scene.key.replace('scene_', ''),
    }));
  }

  _renderHeader(scene, headerMode) {
    if (
      get(this.props, 'screenProps.fakeHeader', false) &&
      get(scene.route, 'params.webView', false)
    ) {
      return this._renderFakeHeader(scene, headerMode);
    }

    const { header } = this._getScreenDetails(scene).options;

    if (typeof header !== 'undefined' && typeof header !== 'function') {
      return header;
    }

    const renderHeader = header || (props => <Header {...props} />);

    // We need to explicitly exclude `mode` since Flow doesn't see
    // mode: headerMode override below and reports prop mismatch
    const { mode, ...passProps } = this.props;

    return renderHeader({
      ...passProps,
      scene,
      webView: this.webView,
      mode: headerMode,
      getScreenDetails: this._getScreenDetails,
    });
  }

  _renderFakeHeader(scene, headerMode) {
    const statusbarHeight = Platform.OS === 'ios' ? 20 : 0;
    const backgroundColor = get(this.props, 'screenProps.statusBar.backgroundColor', '#fff');
    const statusBarStyle = get(this.props, 'screenProps.statusBar.barStyle', 'dark-content');

    const miniHeaderStyle = {
      width: Dimensions.get('window').width,
      height: statusbarHeight,
      backgroundColor,
    };

    const headerStyle = {
      height: 0,
      position: 'absolute',
      width: '100%',
      left: 0,
      overflow: 'hidden',
      top: statusbarHeight,
    };

    const params = get(scene.route, 'params', {});
    const prev = this.getPreviousSceneParams();

    const imgStyle = {
      width: params.webViewWidth || prev.webViewWidth || Dimensions.get('window').width,
      height: params.webViewHeight || prev.webViewHeight || Dimensions.get('window').height - statusbarHeight,
      resizeMode: 'stretch',
    };

    let uri;

    // set image for goBack transition
    // this transition doesnt trigger rerender
    // it modifies the this.fakeHeaderHeight animated property instead
    if (prev) {
      headerStyle.height = this.fakeHeaderHeight;
      uri = prev.image;
    }

    // we are under forward transition
    if (params.webView && params.loading) {
      // if we have an image for the previous screen
      if (prev && prev.headerHeight) {
        headerStyle.height = prev.headerHeight;
        uri = prev.image;
      }

      // if we have an image for current screen
      if (params.headerHeight) {
        headerStyle.height = params.headerHeight;
        uri = params.image;
      }
    }

    return (
      <View>
        <View style={miniHeaderStyle} />
        <StatusBar backgroundColor={backgroundColor} barStyle={statusBarStyle} />
        <Animated.View style={headerStyle}>
          <Image style={imgStyle} source={{ uri }} />
        </Animated.View>
      </View>
    );
  }

  _reset(resetToIndex: number, duration: number): void {
    Animated.timing(this.props.position, {
      toValue: resetToIndex,
      duration,
      easing: Easing.linear(),
      useNativeDriver: this.props.position.__isNative,
    }).start(() => {
      this.hideFakeHeader();
    });
  }

  _goBack(backFromIndex, duration) {
    const { navigation, position, scenes } = this.props;
    const toValue = Math.max(backFromIndex - 1, 0);

    // set temporary index for gesture handler to respect until the action is
    // dispatched at the end of the transition.
    this._immediateIndex = toValue;

    Animated.timing(position, {
      toValue,
      duration,
      easing: Easing.linear(),
      useNativeDriver: position.__isNative,
    }).start(() => {
      this.hideFakeHeader();

      this._immediateIndex = null;
      const backFromScene = scenes.find((s: *) => s.index === toValue + 1);
      if (!this._isResponding && backFromScene) {
        navigation.dispatch(
          NavigationActions.back({ key: backFromScene.route.key }),
        );
      }
    });
  }

  getPreviousSceneParams() {
    const { scenes } = this.props;
    let st = {};
    if (scenes.length > 1) {
      st = scenes[scenes.length - 2].route;
      if (st && st.params) {
        return st.params;
      }
    }

    return {};
  }

  showFakeHeader() {
    const prev = this.getPreviousSceneParams();
    if (prev && prev.headerHeight) {
      this.fakeHeaderHeight.setValue(prev.headerHeight); // FIXME
    }
  }

  hideFakeHeader() {
    this.fakeHeaderHeight.setValue(0);
  }

  render() {
    let floatingHeader = null;
    const headerMode = this._getHeaderMode();
    if (headerMode === 'float') {
      floatingHeader = this._renderHeader(this.props.scene, headerMode);
    }
    const { navigation, position, layout, scene, scenes, mode } = this.props;
    const { index } = navigation.state;
    const isVertical = mode === 'modal';

    const responder = PanResponder.create({
      onPanResponderTerminate: () => {
        this._isResponding = false;
        this._reset(index, 0);
      },
      onPanResponderGrant: () => {
        this.showFakeHeader();
        this.setState({
          webViewScrollEnabled: false,
        });

        position.stopAnimation((value) => {
          this._isResponding = true;
          this._gestureStartValue = value;
        });
      },
      onMoveShouldSetPanResponder: (event, gesture) => {
        if (index !== scene.index) {
          return false;
        }
        const immediateIndex = this._immediateIndex == null
          ? index
          : this._immediateIndex;
        const currentDragDistance = gesture[isVertical ? 'dy' : 'dx'];
        const currentDragPosition =
          event.nativeEvent[isVertical ? 'pageY' : 'pageX'];
        const axisLength = isVertical
          ? layout.height.__getValue()
          : layout.width.__getValue();
        const axisHasBeenMeasured = !!axisLength;

        // Measure the distance from the touch to the edge of the screen
        const screenEdgeDistance = currentDragPosition - currentDragDistance;
        // Compare to the gesture distance relavant to card or modal
        const gestureResponseDistance = isVertical
          ? GESTURE_RESPONSE_DISTANCE_VERTICAL
          : GESTURE_RESPONSE_DISTANCE_HORIZONTAL;
        // GESTURE_RESPONSE_DISTANCE is about 30 or 35. Or 135 for modals
        if (screenEdgeDistance > gestureResponseDistance) {
          // Reject touches that started in the middle of the screen
          return false;
        }

        const hasDraggedEnough =
          Math.abs(currentDragDistance) > RESPOND_THRESHOLD;

        const isOnFirstCard = immediateIndex === 0;
        const shouldSetResponder =
          hasDraggedEnough && axisHasBeenMeasured && !isOnFirstCard;
        return shouldSetResponder;
      },
      onPanResponderMove: (event, gesture) => {
        // Handle the moving touches for our granted responder
        const startValue = this._gestureStartValue;
        const axis = isVertical ? 'dy' : 'dx';
        const axisDistance = isVertical
          ? layout.height.__getValue()
          : layout.width.__getValue();
        const currentValue = I18nManager.isRTL && axis === 'dx'
          ? startValue + gesture[axis] / axisDistance
          : startValue - gesture[axis] / axisDistance;
        const value = clamp(index - 1, currentValue, index);
        position.setValue(value);
      },
      onPanResponderTerminationRequest: () =>
        // Returning false will prevent other views from becoming responder while
        // the navigation view is the responder (mid-gesture)
        false,
      onPanResponderRelease: (event, gesture) => {
        this.setState({
          webViewScrollEnabled: true,
        });
        if (!this._isResponding) {
          return;
        }
        this._isResponding = false;

        const immediateIndex = this._immediateIndex == null
          ? index
          : this._immediateIndex;

        // Calculate animate duration according to gesture speed and moved distance
        const axisDistance = isVertical
          ? layout.height.__getValue()
          : layout.width.__getValue();
        const movedDistance = gesture[isVertical ? 'moveY' : 'moveX'];
        const defaultVelocity = axisDistance / ANIMATION_DURATION;
        const gestureVelocity = gesture[isVertical ? 'vy' : 'vx'];
        const velocity = Math.max(gestureVelocity, defaultVelocity);
        const resetDuration = movedDistance / velocity;
        const goBackDuration = (axisDistance - movedDistance) / velocity;

        // To asyncronously get the current animated value, we need to run stopAnimation:
        position.stopAnimation((value) => {
          // If the speed of the gesture release is significant, use that as the indication
          // of intent
          if (gestureVelocity < -0.5) {
            this._reset(immediateIndex, resetDuration);
            return;
          }
          if (gestureVelocity > 0.5) {
            this._goBack(immediateIndex, goBackDuration);
            return;
          }

          // Then filter based on the distance the screen was moved. Over a third of the way swiped,
          // and the back will happen.
          if (value <= index - POSITION_THRESHOLD) {
            this._goBack(immediateIndex, goBackDuration);
          } else {
            this._reset(immediateIndex, resetDuration);
          }
        });
      },
    });

    const { options } = this._getScreenDetails(scene);
    const gesturesEnabled = typeof options.gesturesEnabled === 'boolean'
      ? options.gesturesEnabled
      : Platform.OS === 'ios';

    const handlers = gesturesEnabled ? responder.panHandlers : {};

    return (
      <View
        {...handlers}
        style={styles.container}
        collapsable={false}
        ref={node => this.view = node}>
        <View style={styles.scenes}>
          {scenes.map(s => this._renderCard(s))}
          {this._renderWebView()}
        </View>
        {floatingHeader}
      </View>
    );
  }

  _onMessage = (event) => {
    if (this.props.screenProps.onMessage) {
      this.props.screenProps.onMessage(event);
    }

    const type = getEventType(event);
    if (type === 'pageshow') {
      this.onPageShow();
    } else if (type === 'pagehide') {
      this.onPageHide();
    }
  }

  onLoad = (event) => {
    if (this.props.screenProps.onLoad) {
      this.props.screenProps.onLoad(event);
    }
  };

  onLoadEnd = (event, { key = '' }) => {
    if (this.props.screenProps.onLoadEnd) {
      this.props.screenProps.onLoadEnd(event);
    }

    const action = NavigationActions.setParams({
      key: key.replace('scene_', ''),
      params: { url: event.nativeEvent.url },
    });
    this.props.navigation.dispatch(action);
  };

  onPageShow = () => {
    this.pageLoading = false;
    const action = NavigationActions.setParams({
      key: this.props.scene.key.replace('scene_', ''),
      params: {
        loading: false,
        hideWebView: false,
      },
    });
    this.props.navigation.dispatch(action);
  };

  onPageHide = () => {
    this.pageLoading = true;
    const action = NavigationActions.setParams({
      key: this.props.scene.key.replace('scene_', ''),
      params: {
        loading: true,
        hideWebView: true,
      },
    });
    this.props.navigation.dispatch(action);
  };

  _renderWebView = () => {
    // console.log(this);
    const { layout, position, scene, scenes, screenProps } = this.props;
    const index = scene.index;
    const prevScene = scenes[index - 1] || null;
    const width = layout.initWidth;

    const opacity = position.interpolate({
      inputRange: [index - 1, index - 0.99, index, index + 0.99, index + 1],
      outputRange: [0, 1, 1, 0.3, 0],
    });

    const translateY = 0;
    const translateX = position.interpolate({
      inputRange: [index - 1, index, index + 1],
      outputRange: [width, 0, -10], // [-width, 0, 10]
    });

    const showWebView = (
      get(scene.route, 'params.webView', false) &&
      !get(scene.route, 'params.loading', false) &&
      !get(scene.route, 'params.hideWebView', false)
      // !get(this.props.navigation.state, 'webViewBackPromise', false)
    );

    const windowDimensions = Dimensions.get('window');
    const { webViewWidth, webViewHeight, webViewX, webViewY } = get(scene.route, 'params', {});

    const fakeHeader = get(this.props, 'screenProps.fakeHeader', false) && get(scene.route, 'params.webView', false);
    let top = webViewY || 0;
    if (!fakeHeader && Platform.OS === 'android') {
      top = Header.HEIGHT;
    }

    const style = [{
      position: 'absolute',
      width: webViewWidth || windowDimensions.width,
      height: webViewHeight || windowDimensions.height,
      top,
      left: showWebView ? webViewX || 0 : windowDimensions.width,
      opacity,
      transform: [{ translateX }, { translateY }],
    }];

    const screenURI = get(screenProps, 'initialWebViewURI', null);
    const configURI = get(this.props.router, 'stackConfig.webViewConfig.initialURI', null);
    // TODO: make all options configurable either via screenProps or stackConfig.webViewConfig

    const source = {
      uri: screenURI || configURI || 'about:blank',
    };

    const clientJs = get(screenProps, 'clientJsURI', null);

    return (
      <Animated.View style={style}>
        <GridWebView
          ref={(node) => { this.webView = node; }}
          style={{ flex: 1 }}
          source={source}
          onNavigationStateChange={event => (
            onNavigationStateChange.apply(this, [event, scene, prevScene])
          )}
          navigation={this.props.navigation}
          clientJs={clientJs}
          onMessage={this._onMessage}
          onLoad={this.onLoad}
          onLoadEnd={event => this.onLoadEnd(event, scene)}
          scrollEnabled={get(this, 'state.webViewScrollEnabled', true)}
        />
      </Animated.View>
    );
  };
}

function getEventType(event) {
  return get(event.body.data, 'type', null);
}

export default CardStack;
